#! /usr/bin/env python3
# Licensed under a 3-clause BSD style license - see LICENSE
from __future__ import print_function
import sys
from optparse import OptionParser

try:
    import pyfits
except ImportError:
    try:
        from astropy.io import fits as pyfits
    except ImportError:
        raise ImportError("Cannot import either pyfits or astropy.io.fits")
from numpy import *

from astrometry.util.fits import *
from astrometry.util.healpix import *
from astrometry.util.starutil_numpy import *

def get_2mass_sources(ra, dec, radius=1, basefn=None):
    twomass_nside = 9

    if basefn is None:
        twomass_pat = '2mass_hp%03i.fits'
    else:
        twomass_pat = basefn

    hps = healpix_rangesearch(ra, dec, radius, twomass_nside)
    print('2MASS healpixes in range:', hps)
    allU = None
    for hp in hps:
        fn = twomass_pat % hp
        print('2MASS filename:', fn)
        U = fits_table(fn)
        print(len(U), 'sources')
        I = (degrees_between(ra, dec, U.ra, U.dec) < radius)
        print('%i 2MASS stars within range.' % sum(I))
        U = U[I]
        if allU is None:
            allU = U
        else:
            allU.append(U)
    return allU

if __name__ == '__main__':
    parser = OptionParser(usage='%prog [options] <ra> <dec> <output-filename>')

    parser.add_option('-r', dest='radius', type='float', help='Search radius, in deg (default 1 deg)')
    parser.add_option('-b', dest='basefn', help='Base filename of 2MASS FITS files (default: 2mass_hp%03i.fits)')
    parser.add_option('-B', dest='band', help='Band (J, H, or K) to use for cuts')
    parser.set_defaults(radius=1.0, basefn=None, band='J')

    (opt, args) = parser.parse_args()
    if len(args) != 3:
        parser.print_help()
        print()
        print('Got extra arguments:', args)
        sys.exit(-1)

    # parse RA,Dec.
    ra = float(args[0])
    dec = float(args[1])
    outfn = args[2]

    band = opt.band.lower()

    # ugh!
    opts = {}
    for k in ['radius', 'basefn']:
        opts[k] = getattr(opt, k)

    X = get_2mass_sources(ra, dec, **opts)
    print('Got %i 2MASS sources.' % len(X))

    #print X.about()

    print('Applying cuts...')
    I = logical_not(X.minor_planet)
    print('not minor planet:', sum(I))
    qual = X.get(band + '_quality')
    # work around dumb bug where it's a single-char column rather than a byte.
    nobrightness = chr(0)
    I = logical_and(I, (qual != nobrightness))
    print('not NO_BRIGHTNESS', sum(I))
    print(len(X))
    print(len(X.j_cc))
    cc = array(X.getcolumn(band + '_cc'))
    ccnone = chr(0)
    #print 'cc shape', cc.shape
    #print cc[:10]
    #print ccnone
    I = logical_and(I, (cc == ccnone))
    print('CC_NONE', sum(I))
    X = X[I]
    print('%i pass cuts' % len(X))

    print('Writing to', outfn)
    X.write_to(outfn)
    



#! /usr/bin/env python3
# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
if __name__ == '__main__':
    import matplotlib
    matplotlib.use('Agg')

import pylab as plt
import numpy as np
import sys
from math import sqrt, floor, ceil
from astrometry.util.util import Sip
from optparse import OptionParser

def plotDistortion(sip, W, H, ncells, exaggerate=1., doclf=True):
    '''
    Produces a plot showing the SIP distortion that was found, by drawing
    a grid and distorting it.  Allows exaggeration of the distortion for ease
    of visualization.

    sip -- an astrometry.util.Sip object
       (duck-type: has "(dx,dy) = sip.get_distortion(x,y)")
    W, H -- the image size
    ncells -- the approximate number of grid cells to split the image into.
    exaggerate -- the factor by which to exaggerate the distortion.
    
    '''
    ncells = float(ncells)
    cellsize = sqrt(W * H / ncells)
    nw = int(floor(W / cellsize))
    nh = int(floor(H / cellsize))
    #print 'Grid cell size', cellsize
    #print 'N cells', nw, 'x', nh
    cx = np.arange(nw+1) * cellsize + ((W - (nw*cellsize))/2.)
    cy = np.arange(nh+1) * cellsize + ((H - (nh*cellsize))/2.)

    # pixel step size for grid lines
    step = 50

    nx = int(np.ceil(W / float(step)))
    ny = int(np.ceil(H / float(step)))
    
    #xx = np.arange(-step, W+2*step, step)
    #yy = np.arange(-step, H+2*step, step)
    xx = np.linspace(1, W, nx)
    yy = np.linspace(1, H, ny)
    
    if doclf:
        plt.clf()

    for y in cy:
        dx,dy = [],[]
        for x in xx:
            dxi,dyi = sip.get_distortion(x, y)
            dx.append(dxi)
            dy.append(dyi)
        plt.plot(xx, y*np.ones_like(xx), 'k-', zorder=10)
        dx = np.array(dx)
        dy = np.array(dy)
        if exaggerate != 1:
            dx += (exaggerate * (dx - xx))
            dy += (exaggerate * (dy - y))
        plt.plot(dx, dy, 'r-', zorder=20)

    for x in cx:
        dx,dy = [],[]
        for y in yy:
            dxi,dyi = sip.get_distortion(x, y)
            dx.append(dxi)
            dy.append(dyi)
        plt.plot(x*np.ones_like(yy), yy, 'k-', zorder=10)
        dx = np.array(dx)
        dy = np.array(dy)
        if exaggerate != 1:
            dx += (exaggerate * (dx - x))
            dy += (exaggerate * (dy - yy))
        plt.plot(dx, dy, 'r-', zorder=20, clip_on=False)
    
    plt.axis('scaled')
    plt.axis([0, W, 0, H])

def plotDistortionFile(sipfn, ext, ncells, **kwargs):
    wcs = Sip(sipfn, ext)
    if wcs is None:
        raise RuntimeError('Failed to open WCS file %s' % sipfn)
    plotDistortion(wcs, wcs.get_width(), wcs.get_height(), ncells, **kwargs)



if __name__ == '__main__':
    parser = OptionParser('usage: %prog [options] <wcs-filename> <plot-output-filename>')
    parser.add_option('-e', dest='ext', type='int', help='FITS extension to read WCS from (default 0)')
    parser.add_option('-x', dest='exaggerate', type='float', help='Exaggeration factor')
    parser.add_option('-c', dest='cells', type='int', help='Approx. number of pieces to cut image into (default:18)')
    parser.set_defaults(ext=0, cells=18, exaggerate=1.)
    opt,args = parser.parse_args()
    if len(args) != 2:
        parser.print_help()
        sys.exit(-1)

    wcsfn = args[0]
    outfn = args[1]

    plotDistortionFile(wcsfn, opt.ext, opt.cells, exaggerate=opt.exaggerate)
    plt.savefig(outfn)



# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
from __future__ import print_function
import matplotlib
matplotlib.use('Agg')

import sys
from optparse import *

import numpy as np
from pylab import *
from numpy import *
#from astrometry.util.sip import *
from astrometry.util.util import *

def plot_distortions(wcsfn, ex=1, ngridx=10, ngridy=10, stepx=10, stepy=10):
    wcs = Sip(wcsfn)
    W,H = wcs.wcstan.imagew, wcs.wcstan.imageh

    xgrid = np.linspace(0, W, ngridx)
    ygrid = np.linspace(0, H, ngridy)
    X = np.linspace(0, W, int(ceil(W/stepx)))
    Y = np.linspace(0, H, int(ceil(H/stepy)))

    xlo,xhi,ylo,yhi = 0,W,0,H

    for x in xgrid:
        DX,DY = [],[]
        xx,yy = [],[]
        UX,UY = [],[]
        for y in Y:
            dx,dy = wcs.get_distortion(x, y)
            ux,uy = wcs.get_undistortion(dx, dy)
            print('x,y', (x,y), 'dx,dy', (dx,dy), 'ux,uy', (ux,uy))
            xx.append(x)
            yy.append(y)
            DX.append(dx)
            DY.append(dy)
            UX.append(ux)
            UY.append(uy)
        DX = np.array(DX)
        DY = np.array(DY)
        UX = np.array(UX)
        UY = np.array(UY)
        xx = np.array(xx)
        yy = np.array(yy)
        EX = DX + ex * (DX - xx)
        EY = DY + ex * (DY - yy)
        plot(xx, yy, 'k-', alpha=0.5)
        plot(EX, EY, 'r-')
        plot(UX, UY, 'b-', alpha=0.5)
        xlo = min(xlo, min(EX))
        xhi = max(xhi, max(EX))
        ylo = min(ylo, min(EY))
        yhi = max(yhi, max(EY))

    for y in ygrid:
        DX,DY = [],[]
        xx,yy = [],[]
        UX,UY = [],[]
        for x in X:
            dx,dy = wcs.get_distortion(x, y)
            ux,uy = wcs.get_undistortion(dx, dy)
            DX.append(dx)
            DY.append(dy)
            UX.append(ux)
            UY.append(uy)
            xx.append(x)
            yy.append(y)
        DX = np.array(DX)
        DY = np.array(DY)
        xx = np.array(xx)
        yy = np.array(yy)
        EX = DX + ex * (DX - xx)
        EY = DY + ex * (DY - yy)
        plot(xx, yy, 'k-', alpha=0.5)
        plot(EX, EY, 'r-')
        plot(UX, UY, 'b-', alpha=0.5)
        xlo = min(xlo, min(EX))
        xhi = max(xhi, max(EX))
        ylo = min(ylo, min(EY))
        yhi = max(yhi, max(EY))

    plot([wcs.wcstan.crpix[0]], [wcs.wcstan.crpix[1]], 'rx')

    #axis([0, W, 0, H])
    axis('scaled')
    axis([xlo,xhi,ylo,yhi])
    #axis('tight')

if __name__ == '__main__':
    parser = OptionParser(usage='%prog [options] <wcs-filename> <plot-filename>')
    parser.add_option('-e', '--ex', '--exaggerate', dest='ex', type='float', help='Exaggerate the distortion by this factor')
    #parser.add_option('-s', '--scale', dest='scale', type='float', help='Scale the
    parser.add_option('-n', dest='nsteps', type='int', help='Number of grid lines to plot')
 
    parser.set_defaults(ex=1.)

    opt,args = parser.parse_args()

    if len(args) != 2:
        parser.print_help()
        sys.exit(-1)

    wcsfn = args[0]
    outfn = args[1]

    args = {}
    if opt.ex is not None:
        args['ex'] = opt.ex
    if opt.nsteps is not None:
        args['ngridx'] = opt.nsteps
        args['ngridy'] = opt.nsteps

    clf()
    plot_distortions(wcsfn, **args)
    tt = 'SIP distortions: %s' % wcsfn
    if opt.ex != 1:
        tt += ' (exaggerated by %g)' % opt.ex
    title(tt)
    savefig(outfn)
    



# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
from __future__ import print_function
import matplotlib
matplotlib.use('Agg')
import pylab as plt

import sys
from astrometry.sdss.dr8 import *
import numpy as np

def test_astrans(sdss, r,c,f,b):
    bandnum = band_index(b)
    sdss.retrieve('frame', r, c, f, b)
    frame = sdss.readFrame(r, c, f, b)
    astrans = frame.getAsTrans()
    sdss.retrieve('photoObj', r, c, f)
    obj = sdss.readPhotoObj(r, c, f)
    tab = obj.getTable()
    #tab.about()
    x,y = tab.colc[:,bandnum], tab.rowc[:,bandnum]
    ra,dec = tab.ra, tab.dec


    for r,d in zip(ra,dec):
        print('ra,dec', r,d)
        #print 'py:'
        x1,y1 = astrans.radec_to_pixel_single_py(r, d)
        print('  py', x1,y1)
        #print 'c:'
        x2,y2 = astrans.radec_to_pixel_single_c(r, d)
        print('  c', x2,y2)
        assert(np.abs(x1 - x2) < 1e-6)
        assert(np.abs(y1 - y2) < 1e-6)




    r2,d2 = astrans.pixel_to_radec(x, y)
    plt.clf()
    plt.plot(ra, dec, 'r.')
    plt.plot(r2, d2, 'bo', mec='b', mfc='none')
    plt.savefig('rd.png')

    r3,d3 = [],[]
    for xi,yi in zip(x,y):
        ri,di = astrans.pixel_to_radec(xi, yi)
        r3.append(ri)
        d3.append(di)
    plt.clf()
    plt.plot(ra, dec, 'r.')
    plt.plot(r3, d3, 'bo', mec='b', mfc='none')
    plt.savefig('rd3.png')

    x2,y2 = astrans.radec_to_pixel(ra, dec)
    plt.clf()
    plt.plot(x, y, 'r.')
    plt.plot(x2, y2, 'bo', mec='b', mfc='none')
    plt.savefig('xy.png')

    x3,y3 = [],[]
    for ri,di in zip(ra, dec):
        xi,yi = astrans.radec_to_pixel(ri, di)
        x3.append(xi)
        y3.append(yi)
    plt.clf()
    plt.plot(x, y, 'r.')
    plt.plot(x3, y3, 'bo', mec='b', mfc='none')
    plt.savefig('xy3.png')


if __name__ == '__main__':
    sdss = DR8()
    #test_astrans(sdss, 4623, 1, 203, 'r')
    test_astrans(sdss, 5065, 1, 68, 'r')
    sys.exit(0)

    fnew  = sdss.readFrame(4623, 1, 203, 'r', filename='frame-r-004623-1-0203.fits')
    print('fnew:', fnew)
    forig = sdss.readFrame(4623, 1, 203, 'r', 'frame-r-004623-1-0203.fits.orig')
    print('forig:', forig)

    frame = sdss.readFrame(3712, 3, 187, 'r')
    print('frame:', frame)
    img = frame.getImage()
    print('  image', img.shape)

    fpobj = sdss.readFpObjc(6581, 2, 135)
    print('fpobj:', fpobj)

    fpm = sdss.readFpM(6581, 2, 135, 'i')
    print('fpm:', fpm)

    psf = sdss.readPsField(6581, 2, 135)
    print('psfield:', psf)


    



# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
from __future__ import print_function
from __future__ import absolute_import

import os
try:
    # py2
    import cPickle as pickle
except:
    import pickle

def trymakedirs(fn, dir=False):
    if dir is True:
        dirnm = os.path.dirname(fn)
    else:
        dirnm = fn
    if not os.path.exists(dirnm):
        try:
            os.makedirs(dirnm)
        except:
            pass

def _filesize(fn):
    st = os.stat(fn)
    return st.st_size

def file_size(fn, error=None):
    if error is None:
        return _filesize(fn)
    try:
        return _filesize(fn)
    except OSError:
        return error

def read_file(fn):
    return open(fn).read()

def write_file(data, fn):
    f = open(fn, 'wb')
    f.write(data)
    f.close()
    
def pickle_to_file(data, fn):
    f = open(fn, 'wb')
    # MAGIC -1: highest pickle protocol
    pickle.dump(data, f, -1)
    f.close()

def unpickle_from_file(fn):
    f = open(fn, 'rb')
    data = pickle.load(f)
    # necessary?
    f.close()
    return data

def get_svn_version():
    from .run_command import run_command
    version = {}
    rtn,out,err = run_command('svn info')
    if rtn != 0:
        import sys
        print('Error getting SVN version: rtn', rtn, '\nOut:', out, '\nErr:', err, file=sys.stderr)
    assert(rtn == 0)
    lines = out.split('\n')
    lines = [l for l in lines if len(l)]
    for l in lines:
        words = l.split(':', 1)
        words = [w.strip() for w in words]
        version[words[0]] = words[1]
    return version

def get_git_version():
    '''
    eg,
    {'commit': 'a5c7865efd188715a8436ef7be23e38448e2aa60', 'describe': 'v1.0'}
    '''
    from .run_command import run_command
    version = {}
    rtn,out,err = run_command('git log --max-count=1 | head -n 1')
    assert(rtn == 0)
    lines = out.split('\n')
    lines = [l for l in lines if len(l)]
    for l in lines:
        words = l.split(' ', 1)
        words = [w.strip() for w in words]
        version[words[0]] = words[1]

    rtn,out,err = run_command('git describe')
    if rtn == 0:
        # this can fail if there has been no "git tag"
        lines = out.split('\n')
        lines = [l for l in lines if len(l)]
        assert(len(lines) == 1)
        version['describe'] = lines[0]

    return version




# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE

# Generates FITS tables from CSV lists of OpenNGC entries and names.

from __future__ import print_function

import csv

from astrometry.util.fits import fits_table
import numpy as np


def convert_openngc_entries():
    entries = []

    with open('openngc-entries.csv') as f:
        for is_ngc, num, ra, dec, size in csv.reader(f, delimiter=';'):
            is_ngc = (is_ngc == '1')
            num = int(num)
            ra = float(ra) if ra else 0.0
            dec = float(dec) if dec else 0.0

            # Convert from diameter in arcmins to radius in degrees.
            radius = float(size) / (2.0 * 60.0) if size else 0.0

            entries.append({
                'is_ngc': is_ngc,
                'ra': ra,
                'dec': dec,
                'radius': radius,
                'num': num,
            })

    T = fits_table()
    for key in ['is_ngc', 'ra', 'dec', 'radius', 'num']:
        T.set(key, [x[key] for x in entries])

    T.to_np_arrays()

    T.name = np.array(['NGC %i' % n if isngc else 'IC %i' % n
                       for n, isngc in zip(T.num, T.is_ngc)])

    for key in ['ra', 'dec', 'radius']:
        T.set(key, T.get(key).astype(np.float32))
    T.num = T.num.astype(np.int16)

    units_dict = {
        'ra': 'deg',
        'dec': 'deg',
        'radius': 'deg',
    }

    NGC = T[T.is_ngc]
    NGC.rename('num', 'ngcnum')
    NGC.delete_column('is_ngc')
    units = [units_dict.get(c, '') for c in NGC.get_columns()]
    NGC.writeto('openngc-ngc.fits', units=units)

    IC = T[np.logical_not(T.is_ngc)]
    IC.rename('num', 'icnum')
    IC.delete_column('is_ngc')
    units = [units_dict.get(c, '') for c in IC.get_columns()]
    IC.writeto('openngc-ic.fits', units=units)


def convert_openngc_names():
    names = []

    with open('openngc-names.csv') as f:
        for is_ngc, num, name in csv.reader(f, delimiter=';'):

            is_ngc = bool(is_ngc)

            num = int(num)

            identifier = '%s%d' % ('' if is_ngc else 'I', num)

            names.append({
                'Object': name,
                'Name': identifier,
            })

    T = fits_table()
    for key in ['Object', 'Name']:
        T.set(key, [x[key] for x in names])
    T.writeto('openngc-names.fits')


if __name__ == '__main__':
    convert_openngc_entries()
    convert_openngc_names()



# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
from __future__ import print_function
import matplotlib
matplotlib.use('Agg')
import pylab as plt
import numpy as np

import emcee
import triangle

from astrometry.util.util import *
from astrometry.util.plotutils import *
from astrometry.util.fits import *
from astrometry.solver.solver import *


class McTweak(object):
    def __init__(self, wcs, xy, rd):
        self.refra  = rd.ra
        self.refdec = rd.dec
        self.testxy = np.vstack((xy.x, xy.y)).T
        nt = len(xy)
        sig2 = 1.
        self.testsig2 = np.zeros(nt) + sig2

        self.W = wcs.get_width()
        self.H = wcs.get_height()
        self.distractors = 0.25
        ## Accept: set to ~inf?
        self.logodds_bail = -1e100
        self.logodds_accept = 1e12

        self.wcs = wcs
        
    def __call__(self, args):
        # plug args into wcs

        # make a local copy...
        wcs = Sip(self.wcs)
        set_sip_args(wcs, args)

        # sip.radec2pixelxy uses the *inverse* SIP polynomials... compute 'em
        sip_compute_inverse_polynomials(wcs, 20, 20, 1, self.W, 1, self.H)
        
        ok,x,y = wcs.radec2pixelxy(self.refra, self.refdec)
        refxy = np.vstack((x,y)).T
        
        logodds = verify_star_lists_np(refxy, self.testxy, self.testsig2,
                                       self.W * self.H, self.distractors,
                                       self.logodds_bail, self.logodds_accept)
        return logodds



def set_sip_args(wcs, args):
    args = list(reversed(args))
    r = args.pop()
    d = args.pop()
    wcs.set_crval((r,d))
    CD = (args.pop(), args.pop(), args.pop(), args.pop())
    wcs.set_cd(CD)

    order = wcs.a_order
    for p in range(0, order+1):
        for q in range(0, order+1-p):
            if p+q <= 1:
                continue
            assert(p + q <= order)
            wcs.set_a_term(p, q, args.pop())

    order = wcs.b_order
    for p in range(0, order+1):
        for q in range(0, order+1-p):
            if p+q <= 1:
                continue
            assert(p + q <= order)
            wcs.set_b_term(p, q, args.pop())

    assert(len(args) == 0)

def get_sip_args(wcs):
    W,H = wcs.get_width(), wcs.get_height()
    S = max(W, H)
    args = []
    sigs = []
    r,d = wcs.get_crval()
    pixscale = wcs.pixel_scale()
    args.extend([r,d])
    sigs.extend([pixscale/3600.]*2)

    cd1,cd2,cd3,cd4 = wcs.get_cd()
    args.extend([cd1,cd2,cd3,cd4])
    sigs.extend([max(x/1000., pixscale/3600./S) for x in [cd1,cd2,cd3,cd4]])
    
    order = wcs.a_order
    for p in range(0, order+1):
        for q in range(0, order+1-p):
            if p+q <= 1:
                continue
            assert(p + q <= order)
            args.append(wcs.get_a_term(p, q))
            sigs.append(S**-(p+q))
    order = wcs.b_order
    for p in range(0, order+1):
        for q in range(0, order+1-p):
            if p+q <= 1:
                continue
            assert(p + q <= order)
            args.append(wcs.get_b_term(p, q))
            sigs.append(S**-(p+q))
    return args, sigs


def mctweak(wcs, xy, rd):
    obj = McTweak(wcs, xy, rd)

    # Initial args
    args,sigs = get_sip_args(wcs)

    print('Args:', args)
    print('Sigs:', sigs)
    print('Number of arguments:', len(args))
    print('Logodds:', obj(args))

    ndim, nwalkers = len(args), 100
    p0 = emcee.utils.sample_ball(args, sigs, size=nwalkers)
    print('p0', p0.shape)

    ps = PlotSequence('mctweak')

    W,H = wcs.get_width(), wcs.get_height()
    mywcs = Sip(wcs)
    
    sampler = emcee.EnsembleSampler(nwalkers, ndim, obj)
    lnp0, rstate = None, None
    pp = []
    for step in range(10000):
        print('Step', step)
        p0,lnp0,rstate = sampler.run_mcmc(p0, 1, lnprob0=lnp0, rstate0=rstate)
        print('Best logprob:', np.max(lnp0))
        i = np.argmax(lnp0)
        print('Best args:', p0[i,:])

        pp.extend(sampler.flatchain)
        sampler.reset()
        
        if step % 100 != 0:
            continue

        
        plt.clf()
        plt.plot(obj.testxy[:,0], obj.testxy[:,1], 'r.')
        for args in p0[np.random.permutation(nwalkers)[:10],:]:
            set_sip_args(mywcs, args)
            sip_compute_inverse_polynomials(mywcs, 20, 20, 1, W, 1, H)
            ok,x,y = mywcs.radec2pixelxy(obj.refra, obj.refdec)
            plt.plot(x, y, 'bo', mec='b', mfc='none', alpha=0.25)

            ex = 10.
            ngridx = ngridy = 10
            stepx = stepy = 100
            xgrid = np.linspace(0, W, ngridx)
            ygrid = np.linspace(0, H, ngridy)
            X = np.linspace(0, W, int(np.ceil(W/stepx)))
            Y = np.linspace(0, H, int(np.ceil(H/stepy)))
            for x in xgrid:
                DX,DY = [],[]
                xx,yy = [],[]
                for y in Y:
                    dx,dy = mywcs.get_distortion(x, y)
                    xx.append(x)
                    yy.append(y)
                    DX.append(dx)
                    DY.append(dy)
                DX = np.array(DX)
                DY = np.array(DY)
                xx = np.array(xx)
                yy = np.array(yy)
                EX = DX + ex * (DX - xx)
                EY = DY + ex * (DY - yy)
                #plot(xx, yy, 'k-', alpha=0.5)
                plt.plot(EX, EY, 'b-', alpha=0.1)

            for y in ygrid:
                DX,DY = [],[]
                xx,yy = [],[]
                for x in X:
                    dx,dy = mywcs.get_distortion(x, y)
                    DX.append(dx)
                    DY.append(dy)
                    xx.append(x)
                    yy.append(y)
                DX = np.array(DX)
                DY = np.array(DY)
                xx = np.array(xx)
                yy = np.array(yy)
                EX = DX + ex * (DX - xx)
                EY = DY + ex * (DY - yy)
                #plot(xx, yy, 'k-', alpha=0.5)
                plt.plot(EX, EY, 'b-', alpha=0.1)
                
        for x in xgrid:
            plt.plot(x+np.zeros_like(Y), Y, 'k-', alpha=0.5)
        for y in ygrid:
            plt.plot(X, y+np.zeros_like(X), 'k-', alpha=0.5)
                
        plt.axis([1, W, 1, H])
        plt.axis('scaled')
        ps.savefig()

        pp = np.vstack(pp)
        print('pp', pp.shape)
        
        # plt.clf()
        # triangle.corner(pp, plot_contours=False)
        # ps.savefig()

        pp = []

wcs = Tan('bok-01.wcs', 0)
sip = Sip(wcs)

# sip.a_order = 3
# sip.b_order = 3
# sip.ap_order = 4
# sip.bp_order = 4
sip.a_order = 2
sip.b_order = 2
sip.ap_order = 3
sip.bp_order = 3

xy = fits_table('bok-01.axy')
rd = fits_table('bok-01.rdls')

mctweak(sip, xy, rd)




# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
from __future__ import print_function
import os
from astrometry.util.fits import fits_table
from astrometry.util.miscutils import get_overlapping_region
import numpy as np
from functools import reduce

try:
    import cutils
except:
    cutils = None

cas_flags = dict(
    CANONICAL_CENTER = 0x0000000000000001,
    BRIGHT = 0x0000000000000002,
    EDGE = 0x0000000000000004,
    BLENDED = 0x0000000000000008,
    CHILD = 0x0000000000000010,
    PEAKCENTER = 0x0000000000000020,
    NODEBLEND = 0x0000000000000040,
    NOPROFILE = 0x0000000000000080,
    NOPETRO = 0x0000000000000100,
    MANYPETRO = 0x0000000000000200,
    NOPETRO_BIG = 0x0000000000000400,
    DEBLEND_TOO_MANY_PEAKS = 0x0000000000000800,
    COSMIC_RAY = 0x0000000000001000,
    MANYR50 = 0x0000000000002000,
    MANYR90 = 0x0000000000004000,
    BAD_RADIAL = 0x0000000000008000,
    INCOMPLETE_PROFILE = 0x0000000000010000,
    INTERP = 0x0000000000020000,
    SATURATED = 0x0000000000040000,
    NOTCHECKED = 0x0000000000080000,
    SUBTRACTED = 0x0000000000100000,
    NOSTOKES = 0x0000000000200000,
    BADSKY = 0x0000000000400000,
    PETROFAINT = 0x0000000000800000,
    TOO_LARGE = 0x0000000001000000,
    DEBLENDED_AS_PSF = 0x0000000002000000,
    DEBLEND_PRUNED = 0x0000000004000000,
    ELLIPFAINT = 0x0000000008000000,
    BINNED1 = 0x0000000010000000,
    BINNED2 = 0x0000000020000000,
    BINNED4 = 0x0000000040000000,
    MOVED = 0x0000000080000000,
    DEBLENDED_AS_MOVING = 0x0000000100000000,
    NODEBLEND_MOVING = 0x0000000200000000,
    TOO_FEW_DETECTIONS = 0x0000000400000000,
    BAD_MOVING_FIT = 0x0000000800000000,
    STATIONARY = 0x0000001000000000,
    PEAKS_TOO_CLOSE = 0x0000002000000000,
    MEDIAN_CENTER = 0x0000004000000000,
    LOCAL_EDGE = 0x0000008000000000,
    BAD_COUNTS_ERROR = 0x0000010000000000,
    BAD_MOVING_FIT_CHILD = 0x0000020000000000,
    DEBLEND_UNASSIGNED_FLUX = 0x0000040000000000,
    SATUR_CENTER = 0x0000080000000000,
    INTERP_CENTER = 0x0000100000000000,
    DEBLENDED_AT_EDGE = 0x0000200000000000,
    DEBLEND_NOPEAK = 0x0000400000000000,
    PSF_FLUX_INTERP = 0x0000800000000000,
    TOO_FEW_GOOD_DETECTIONS = 0x0001000000000000,
    CENTER_OFF_AIMAGE = 0x0002000000000000,
    DEBLEND_DEGENERATE = 0x0004000000000000,
    BRIGHTEST_GALAXY_CHILD = 0x0008000000000000,
    CANONICAL_BAND = 0x0010000000000000,
    AMOMENT_FAINT = 0x0020000000000000,
    AMOMENT_SHIFT = 0x0040000000000000,
    AMOMENT_MAXITER = 0x0080000000000000,
    MAYBE_CR = 0x0100000000000000,
    MAYBE_EGHOST = 0x0200000000000000,
    NOTCHECKED_CENTER = 0x0400000000000000,
    OBJECT2_HAS_SATUR_DN = 0x0800000000000000,
    OBJECT2_DEBLEND_PEEPHOLE = 0x1000000000000000,
    GROWN_MERGED = 0x2000000000000000,
    HAS_CENTER = 0x4000000000000000,
    RESERVED = 0x8000000000000000,
    )

# From:
# http://www.sdss3.org/svn/repo/idlutils/trunk/data/sdss/sdssMaskbits.par
# via
#     s = open('sdssMaskbits.par').read()
#     bits = []
#     for line in s.split('\n'):
#       sp = line.split()
#       line = (int(sp[2]), sp[3], ' '.join(sp[4:]))
#       bits.append(line)
#       print repr(bits).replace('), ', '),\n ')
#
photo_flags1_info = [
    # masktype OBJECT1 32 "Object flags from photo reductions for SDSS (first 32)"
    (0, 'CANONICAL_CENTER', '"The quantities (psf counts, model fits and likelihoods) that are usually determined at an object\'s center as determined band-by-band were in fact determined at the canonical center (suitably transformed). This is due to the object being to close to the edge to extract a profile at the local center, and OBJECT1_EDGE is also set."'),
    (1, 'BRIGHT', '"Indicates that the object was detected as a bright object. Since these are typically remeasured as faint objects, most users can ignore BRIGHT objects."'),
    (2, 'EDGE', '"Object is too close to edge of frame in this band."'),
    (3, 'BLENDED', '"Object was determined to be a blend. The flag is set if: more than one peak is detected within an object in a single band together; distinct peaks are found when merging different colours of one object together; or distinct peaks result when merging different objects together. "'),
    (4, 'CHILD', '"Object is a child, created by the deblender."'),
    (5, 'PEAKCENTER', '"Given center is position of peak pixel, as attempts to determine a better centroid failed."'),
    (6, 'NODEBLEND', '"Although this object was marked as a blend, no deblending was attempted."'),
    (7, 'NOPROFILE', '"Frames couldn\'t extract a radial profile."'),
    (8, 'NOPETRO', '" No Petrosian radius or other Petrosian quanties could be measured."'),
    (9, 'MANYPETRO', '"Object has more than one possible Petrosian radius."'),
    (10, 'NOPETRO_BIG', '"The Petrosian ratio has not fallen to the value at which the Petrosian radius is defined at the outermost point of the extracted radial profile. NOPETRO is set, and the Petrosian radius is set to the outermost point in the profile."'),
    (11, 'DEBLEND_TOO_MANY_PEAKS', '"The object had the OBJECT1_DEBLEND flag set, but it contained too many candidate children to be fully deblended. This flag is only set in the parent, i.e. the object with too many peaks."'),
    (12, 'CR', '"Object contains at least one pixel which was contaminated by a cosmic ray. The OBJECT1_INTERP flag is also set. This flag does not mean that this object is a cosmic ray; rather it means that a cosmic ray has been removed. "'),
    (13, 'MANYR50', '" More than one radius was found to contain 50% of the Petrosian flux. (For this to happen part of the radial profile must be negative)."'),
    (14, 'MANYR90', '"More than one radius was found to contain 90% of the Petrosian flux. (For this to happen part of the radial profile must be negative)."'),
    (15, 'BAD_RADIAL', '" Measured profile includes points with a S/N <= 0. In practice this flag is essentially meaningless."'),
    (16, 'INCOMPLETE_PROFILE', '"A circle, centerd on the object, of radius the canonical Petrosian radius extends beyond the edge of the frame. The radial profile is still measured from those parts of the object that do lie on the frame."'),
    (17, 'INTERP', '" The object contains interpolated pixels (e.g. cosmic rays or bad columns)."'),
    (18, 'SATUR', '"The object contains saturated pixels; INTERP is also set."'),
    (19, 'NOTCHECKED', '"Object includes pixels that were not checked for peaks, for example the unsmoothed edges of frames, and the cores of subtracted or saturated stars."'),
    (20, 'SUBTRACTED', '"Object (presumably a star) had wings subtracted."'),
    (21, 'NOSTOKES', '"Object has no measured Stokes parameters."'),
    (22, 'BADSKY', '"The estimated sky level is so bad that the central value of the radial profile is crazily negative; this is usually the result of the subtraction of the wings of bright stars failing."'),
    (23, 'PETROFAINT', '"At least one candidate Petrosian radius occured at an unacceptably low surface brightness."'),
    (24, 'TOO_LARGE', '" The object is (as it says) too large. Either the object is still detectable at the outermost point of the extracted radial profile (a radius of approximately 260 arcsec), or when attempting to deblend an object, at least one child is larger than half a frame (in either row or column)."'),
    (25, 'DEBLENDED_AS_PSF', '"When deblending an object, in this band this child was treated as a PSF."'),
    (26, 'DEBLEND_PRUNED', '"When solving for the weights to be assigned to each child the deblender encountered a nearly singular matrix, and therefore deleted at least one of them."'),
    (27, 'ELLIPFAINT', '"No isophotal fits were performed."'),
    (28, 'BINNED1', '"The object was detected in an unbinned image."'),
    (29, 'BINNED2', '" The object was detected in a 2x2 binned image after all unbinned detections have been replaced by the background level."'),
    (30, 'BINNED4', '"The object was detected in a 4x4 binned image. The objects detected in the 2x2 binned image are not removed before doing this."'),
    (31, 'MOVED', '"The object appears to have moved during the exposure. Such objects are candidates to be deblended as moving objects."'),
]

photo_flags2_info = [
  (0, 'DEBLENDED_AS_MOVING', '"The object has the MOVED flag set, and was deblended on the assumption that it was moving."'),
  (1, 'NODEBLEND_MOVING', '"The object has the MOVED flag set, but was not deblended as a moving object."'),
  (2, 'TOO_FEW_DETECTIONS', '"The object has the MOVED flag set, but has too few detection to be deblended as moving."'),
  (3, 'BAD_MOVING_FIT', '"The fit to the object as a moving object is too bad to be believed."'),
  (4, 'STATIONARY', '"A moving objects velocity is consistent with zero"'),
  (5, 'PEAKS_TOO_CLOSE', '"Peaks in object were too close (set only in parent objects)."'),
  (6, 'BINNED_CENTER', '"When centroiding the object the object\'s size is larger than the (PSF) filter used to smooth the image."'),
  (7, 'LOCAL_EDGE', '"The object\'s center in some band was too close to the edge of the frame to extract a profile."'),
  (8, 'BAD_COUNTS_ERROR', '"An object containing interpolated pixels had too few good pixels to form a reliable estimate of its error"'),
  (9, 'BAD_MOVING_FIT_CHILD', '"A putative moving child\'s velocity fit was too poor, so it was discarded, and the parent was not deblended as moving"'),
  (10, 'DEBLEND_UNASSIGNED_FLUX', '"After deblending, the fraction of flux assigned to none of the children was too large (this flux is then shared out as described elsewhere)."'),
  (11, 'SATUR_CENTER', '"An object\'s center is very close to at least one saturated pixel; the object may well be causing the saturation."'),
  (12, 'INTERP_CENTER', '"An object\'s center is very close to at least one interpolated pixel."'),
  (13, 'DEBLENDED_AT_EDGE', '"An object so close to the edge of the frame that it would not ordinarily be deblended has been deblended anyway. Only set for objects large enough to be EDGE in all fields/strips."'),
  (14, 'DEBLEND_NOPEAK', '"A child had no detected peak in a given band, but we centroided it anyway and set the BINNED1"'),
  (15, 'PSF_FLUX_INTERP', '"The fraction of light actually detected (as opposed to guessed at by the interpolator) was less than some number (currently 80%) of the total."'),
  (16, 'TOO_FEW_GOOD_DETECTIONS', '"A child of this object had too few good detections to be deblended as moving."'),
  (17, 'CENTER_OFF_AIMAGE', '"At least one peak\'s center lay off the atlas image in some band. This can happen when the object\'s being deblended as moving, or if the astrometry is badly confused."'),
  (18, 'DEBLEND_DEGENERATE', '"At least one potential child has been pruned because its template was too similar to some other child\'s template."'),
  (19, 'BRIGHTEST_GALAXY_CHILD', '"This is the brightest child galaxy in a blend."'),
  (20, 'CANONICAL_BAND', '"This band was the canonical band. This is the band used to measure the Petrosian radius used to calculate the Petrosian counts in each band, and to define the model used to calculate model colors; it has no effect upon the coordinate system used for the OBJC center."'),
  (21, 'AMOMENT_UNWEIGHTED', '"`Adaptive\' moments are actually unweighted."'),
  (22, 'AMOMENT_SHIFT', '"Object\'s center moved too far while determining adaptive moments. In this case, the M_e1 and M_e2 give the (row, column) shift, not the object\'s shape."'),
  (23, 'AMOMENT_MAXITER', '"Too many iterations while determining adaptive moments."'),
  (24, 'MAYBE_CR', '"This object may be a cosmic ray. This bit can get set in the cores of bright stars, and is quite likely to be set for the cores of saturated stars."'),
  (25, 'MAYBE_EGHOST', '"Object appears in the right place to be an electronics ghost."'),
  (26, 'NOTCHECKED_CENTER', '"Center of object lies in a NOTCHECKED region. The object is almost certainly bogus."'),
  (27, 'HAS_SATUR_DN', '"This object is saturated in this band and the bleed trail doesn\'t touch the edge of the frame, we we\'ve made an attempt to add up all the flux in the bleed trails, and to include it in the object\'s photometry. "'),
  (28, 'DEBLEND_PEEPHOLE', '"The deblend was modified by the optimizer"'),
  (29, 'SPARE3', '""'),
  (30, 'SPARE2', '""'),
  (31, 'SPARE1', '""'),
]


specobj_boss_target1_info = [
    # masktype BOSS_TARGET1 64 "BOSS survey primary target selection flags"
    # galaxies
    (0, 'GAL_LOZ', "low-z lrgs"),
    (1, 'GAL_CMASS', "dperp > 0.55, color-mag cut"),
    (2, 'GAL_CMASS_COMM', "dperp > 0.55, commissioning color-mag cut"),
    (3, 'GAL_CMASS_SPARSE', "GAL_CMASS_COMM & (!GAL_CMASS) & (i < 19.9) sparsely sampled"),
    (6, 'SDSS_KNOWN', "Matches a known SDSS spectra"),
    (7, 'GAL_CMASS_ALL', "GAL_CMASS and the entire sparsely sampled region"),
    (8, 'GAL_IFIBER2_FAINT', "ifiber2 > 21.5, extinction corrected. Used after Nov 2010"),
    # galaxies deprecated
    #maskbits BOSS_TARGET1 3 GAL_GRRED "red in g-r"
    #maskbits BOSS_TARGET1 4 GAL_TRIANGLE "GAL_HIZ and !GAL_CMASS"
    #maskbits BOSS_TARGET1 5 GAL_LODPERP "Same as hiz but between dperp00 and dperp0"
    # qsos (1)
    (10, 'QSO_CORE', "restrictive qso selection: commissioning only"),
    (11, 'QSO_BONUS', "permissive qso selection: commissioning only"),
    (12, 'QSO_KNOWN_MIDZ', "known qso between [2.2,9.99]"),
    (13, 'QSO_KNOWN_LOHIZ', "known qso outside of miz range. never target"),
    (14, 'QSO_NN', "Neural Net that match to sweeps/pass cuts"),
    (15, 'QSO_UKIDSS', "UKIDSS stars that match sweeps/pass flag cuts"),
    (16, 'QSO_KDE_COADD', "kde targets from the stripe82 coadd"),
    (17, 'QSO_LIKE', "likelihood method"),
    (18, 'QSO_FIRST_BOSS', "FIRST radio match"),
    (19, 'QSO_KDE', "selected by kde+chi2"),
    # standards
    (20, 'STD_FSTAR', "standard f-stars"),
    (21, 'STD_WD', "white dwarfs"),
    (22, 'STD_QSO', "qso"),
    # template stars
    (32, 'TEMPLATE_GAL_PHOTO', "galaxy templates"),
    (33, 'TEMPLATE_QSO_SDSS1', "QSO templates"),
    (34, 'TEMPLATE_STAR_PHOTO', "stellar templates"),
    (35, 'TEMPLATE_STAR_SPECTRO', "stellar templates (spectroscopically known)"),
    # qsos (2)
    (40, 'QSO_CORE_MAIN', "Main survey core sample"),
    (41, 'QSO_BONUS_MAIN', "Main survey bonus sample"),
    (42, 'QSO_CORE_ED', "Extreme Deconvolution in Core"),
    (43, 'QSO_CORE_LIKE', "Likelihood that make it into core"),
    (44, 'QSO_KNOWN_SUPPZ', "known qso between [1.8,2.15]"),
]


specobj_boss_target1_map = dict([(nm, 1<<bit)
                                 for bit,nm,desc in specobj_boss_target1_info])

photo_flags1_map = dict([(nm, 1<<bit)
                         for bit,nm,desc in photo_flags1_info])
photo_flags2_map = dict([(nm, 1<<bit)
                         for bit,nm,desc in photo_flags2_info])


def band_names():
    return ['u','g','r','i','z']

def band_name(b):
    if b in band_names():
        return b
    if b in [0,1,2,3,4]:
        return 'ugriz'[b]
    raise Exception('Invalid SDSS band: "' + str(b) + '"')

def band_index(b):
    if b in band_names():
        return 'ugriz'.index(b)
    if b in [0,1,2,3,4]:
        return b
    raise Exception('Invalid SDSS band: "' + str(b) + '"')

class SdssDR(object):
    def __init__(self, curl=False, basedir=None):
        self.curl = curl
        self.basedir = basedir
        self.filenames = {}

    def getDRNumber(self):
        return -1
        
    def getFilename(self, filetype, *args, **kwargs):
        for k,v in zip(['run', 'camcol', 'field', 'band'], args):
            kwargs[k] = v
        # convert band number to band character.
        if 'band' in kwargs and kwargs['band'] is not None:
            kwargs['band'] = band_name(kwargs['band'])
        if not filetype in self.filenames:
            return None
        pat = self.filenames[filetype]
        if kwargs.get('rerun', None) is None:
            run = kwargs.get('run', None)
            rerun = self.get_rerun(run)
            kwargs.update(rerun=rerun)
        fn = pat % kwargs
        return fn

    def get_rerun(self, run, field=None):
        return None

    def getPath(self, *args, **kwargs):
        fn = self.getFilename(*args, **kwargs)
        if fn is None:
            return None
        if self.basedir is not None:
            fn = os.path.join(self.basedir, fn)
        return fn

    def setBasedir(self, dirnm):
        self.basedir = dirnm

    def _open(self, fn):

        if self.basedir is not None:
            path = os.path.join(self.basedir, fn)
        else:
            path = fn

        try:
            import fitsio
            return fitsio_wrapper(fitsio.FITS(path))
        except ImportError:
            pass

        try:
            import pyfits
        except ImportError:
            try:
                from astropy.io import fits as pyfits
            except ImportError:
                raise ImportError("Cannot import either pyfits or astropy.io.fits")
        return pyfits.open(path)

class fitsio_wrapper(object):
    def __init__(self, F):
        self.F = F
    def __getitem__(self, k):
        hdu = self.F[k]
        hdu.data = hdu
        return hdu
    
class SdssFile(object):
    def __init__(self, run=None, camcol=None, field=None, band=None, rerun=None,
                 **kwargs):
        '''
        band: string ('u', 'g', 'r', 'i', 'z')
        '''
        self.run = run
        self.camcol = camcol
        self.field = field
        if band is not None:
            self.band = band_name(band)
            self.bandi = band_index(band)
        if rerun is not None:
            self.rerun = rerun
        self.filetype = 'unknown'

    def getRun(self):
        return self.__dict__.get('run', 0)
    def getCamcol(self):
        return self.__dict__.get('camcol', 0)
    def getField(self):
        return self.__dict__.get('field', 0)

    def __str__(self):
        s = 'SDSS ' + self.filetype
        s += ' %i-%i-%i' % (self.getRun(), self.getCamcol(), self.getField())
        if hasattr(self, 'band'):
            s += '-%s' % self.band
        return s


def munu_to_radec_rad(mu, nu, node, incl):
    '''
    Converts SDSS survey coords (mu,nu) into RA,Dec.

    This function requires mu, nu, node, incl to be in RADIANS.

    See munu_to_radec_deg for DEGREES.
    '''
    ra = node + np.arctan2(np.sin(mu - node) * np.cos(nu) * np.cos(incl) -
                           np.sin(nu) * np.sin(incl),
                           np.cos(mu - node) * np.cos(nu))
    dec = np.arcsin(np.sin(mu - node) * np.cos(nu) * np.sin(incl) +
                    np.sin(nu) * np.cos(incl))
    return ra,dec

def munu_to_radec_deg(mu, nu, node, incl):
    '''
    Converts SDSS survey coords (mu,nu) into RA,Dec.

    This function requires mu, nu, node, incl to be in DEGREES.

    See munu_to_radec_rad for RADIANS.
    '''
    mu, nu = np.deg2rad(mu), np.deg2rad(nu)
    node, incl = np.deg2rad(node), np.deg2rad(incl)
    ra,dec = munu_to_radec_rad(mu, nu, node, incl)
    ra, dec = np.rad2deg(ra), np.rad2deg(dec)
    ra += (360. * (ra < 0))
    ra -= (360. * (ra > 360))
    return (ra, dec)


# makes an SDSS AsTrans WCS object look like an anwcs /  Tan / Sip
class AsTransWrapper(object):
    def __init__(self, wcs, w, h, x0=0, y0=0):
        self.wcs = wcs
        self.imagew = w
        self.imageh = h
        self.x0 = x0
        self.y0 = y0
    def pixelxy2radec(self, x, y):
        r,d = self.wcs.pixel_to_radec(x+self.x0-1, y+self.y0-1)
        return r, d
    def radec2pixelxy(self, ra, dec):
        x,y = self.wcs.radec_to_pixel(ra, dec)
        return True, x-self.x0+1, y-self.y0+1

class AsTrans(SdssFile):
    '''
    In DR7, asTrans structures can appear in asTrans files (for a
    whole run) or in tsField files (in astrom/ or fastrom/).

    http://www.sdss.org/dr7/dm/flatFiles/asTrans.html

    In DR8, they are in asTrans files, or in the "frames".

    http://data.sdss3.org/datamodel/files/PHOTO_REDUX/RERUN/RUN/astrom/asTrans.html
    '''
    def __init__(self, *args, **kwargs):
        '''
        node, incl: in radians

        astrans: must be an object with fields:
         {a,b,c,d,e,f}[band]
         {ricut}[band]
         {drow0, drow1, drow2, drow3, dcol0, dcol1, dcol2, dcol3}[band]
         {csrow, cscol, ccrow, cccol}[band]

        cut_to_band: in DR8 frames files, the astrans elements are not arrays;
        in DR7 tsField files they are.

        Note about units in this class:

        mu,nu are in degrees (great circle coords)

        a,d are in degrees (mu0, nu0)
        b,c,e,f are in degrees/pixel (dmu,dnu/drow,dcol)
        drow0,dcol0 are in pixels (distortion coefficients order 0); dpixels
        drow1,dcol1 are unitless dpixels / pixel (distortion coefficients order 1)
        drow2,dcol2 are in 1/pixels (dpixels/pixel**2) (distortion coefficients order 2)
        drow3,dcol3 are in 1/pixels**2 (dpixels/pixel**3) (distortion coefficients order 3)
        csrow,cscol are in pixels/mag (color-dependent shift)
        ccrow,cccol are in pixels (non-color-dependent shift)
        '''
        super(AsTrans, self).__init__(*args, **kwargs)
        self.filetype = 'asTrans'
        self.node = kwargs.get('node', None)
        self.incl = kwargs.get('incl', None)
        astrans = kwargs.get('astrans', None)
        self.trans = {}
        cut = kwargs.get('cut_to_band', True)
        if astrans is not None and hasattr(self, 'bandi'):
            for f in ['a','b','c','d','e','f', 'ricut',
                      'drow0', 'drow1', 'drow2', 'drow3',
                      'dcol0', 'dcol1', 'dcol2', 'dcol3',
                      'csrow', 'cscol', 'ccrow', 'cccol']:
                try:
                    if hasattr(astrans, f):
                        el = getattr(astrans, f)
                        if cut:
                            el = el[self.bandi]
                        self.trans[f] = el
                except:
                    print('failed to get astrans.' + f)
                    import traceback
                    traceback.print_exc()
                    pass

        self._cache_vals()

    @staticmethod
    def read(fn, F=None, primhdr=None, table=None):
        '''
        F: fitsio.FITS object to use an already-open file.
        primhdr: FITS header object for the primary HDU.
        table: astrometry.util.fits table
        '''
        if F is None:
            import fitsio
            F = fitsio.FITS(fn)
        if primhdr is None:
            primhdr = F[0].read_header()
        band   = primhdr['FILTER'].strip()
        run    = primhdr['RUN']
        camcol = primhdr['CAMCOL']
        field  = 0  # 'FRAME' != field
        if table is None:
            tab = fits_table(F[3].read(lower=True))
        else:
            tab = table
        assert(len(tab) == 1)
        tab = tab[0]
        return AsTrans(run, camcol, field, band,
                       node=np.deg2rad(tab.node),
                       incl=np.deg2rad(tab.incl),
                       astrans=tab, cut_to_band=False)
        
    def __str__(self):
        return (SdssFile.__str__(self) +
                ' (node=%g, incl=%g)' % (self.node, self.incl))

    def _cache_vals(self):
        a, b, c, d, e, f = self._get_abcdef()
        determinant = b * f - c * e
        B =  f / determinant
        C = -c / determinant
        E = -e / determinant
        F =  b / determinant
        py,px, qy,qx = self._get_cscc()
        g0, g1, g2, g3 = self._get_drow()
        h0, h1, h2, h3 = self._get_dcol()
        color0 = self._get_ricut()
        self._cached = [self.node, self.incl,
                        a,b,c,d,e,f, B,C,E,F, px,py,qx,qy, g0,g1,g2,g3,
                        h0,h1,h2,h3, color0]

    def _get_abcdef(self):
        return tuple(self.trans[x] for x in 'abcdef')

    def _get_drow(self):
        return tuple(self.trans[x] for x in ['drow0', 'drow1', 'drow2', 'drow3'])

    def _get_dcol(self):
        return tuple(self.trans[x] for x in ['dcol0', 'dcol1', 'dcol2', 'dcol3'])

    def _get_cscc(self):
        return tuple(self.trans[x] for x in ['csrow', 'cscol', 'ccrow', 'cccol'])

    def _get_ricut(self):
        return self.trans['ricut']

    def cd_at_pixel(self, x, y, color=0):
        '''
        (x,y) to numpy array (2,2) -- the CD matrix at pixel x,y:

        [ [ dRA/dx * cos(Dec), dRA/dy * cos(Dec) ],
          [ dDec/dx          , dDec/dy           ] ]

        in FITS these are called:
        [ [ CD11             , CD12              ],
          [ CD21             , CD22              ] ]

          Note: these statements have not been verified by the FDA.
        '''
        ra0,dec0 = self.pixel_to_radec(x, y, color)
        step = 10. # pixels
        rax,decx = self.pixel_to_radec(x+step, y, color)
        ray,decy = self.pixel_to_radec(x, y+step, color)
        cosd = np.cos(np.deg2rad(dec0))
        return np.array([ [ (rax-ra0)/step * cosd, (ray-ra0)/step * cosd ],
                          [ (decx-dec0)/step     , (decy-dec0)/step      ] ])

    def pixel_to_radec(self, x, y, color=0):
        mu, nu = self.pixel_to_munu(x, y, color)
        return self.munu_to_radec(mu, nu)

    def radec_to_pixel_single_py(self, ra, dec, color=0):
        '''RA,Dec -> x,y for scalar RA,Dec.'''
        # RA,Dec -> mu,nu -> prime -> pixel
        mu, nu = self.radec_to_munu_single(ra, dec)
        return self.munu_to_pixel_single(mu, nu, color)

    def radec_to_pixel_single_c(self, ra, dec):
        return cutils.radec_to_pixel(float(ra), float(dec), self._cached)

    def radec_to_pixel(self, ra, dec, color=0):
        mu, nu = self.radec_to_munu(ra, dec)
        return self.munu_to_pixel(mu, nu, color)
    
    def munu_to_pixel(self, mu, nu, color=0):
        xprime, yprime = self.munu_to_prime(mu, nu, color)
        return self.prime_to_pixel(xprime, yprime, color=color)

    munu_to_pixel_single = munu_to_pixel

    def munu_to_prime(self, mu, nu, color=0):
        '''
        mu = a + b * rowm + c * colm
        nu = d + e * rowm + f * colm

        So

        [rowm; colm] = [b,c; e,f]^-1 * [mu-a; nu-d]

        [b,c; e,f]^1 = [B,C; E,F] in the code below, so

        [rowm; colm] = [B,C; E,F] * [mu-a; nu-d]

        '''
        a, b, c, d, e, f = self._get_abcdef()
        determinant = b * f - c * e
        B =  f / determinant
        C = -c / determinant
        E = -e / determinant
        F =  b / determinant
        mua = mu - a
        # in field 6955, g3, 809 we see a~413
        #if mua < -180.:
        #   mua += 360.
        mua += 360. * (mua < -180.)
        yprime = B * mua + C * (nu - d)
        xprime = E * mua + F * (nu - d)
        return xprime,yprime

    def pixel_to_munu(self, x, y, color=0):
        (xprime, yprime) = self.pixel_to_prime(x, y, color)
        a, b, c, d, e, f = self._get_abcdef()
        mu = a + b * yprime + c * xprime
        nu = d + e * yprime + f * xprime
        return (mu, nu)

    def pixel_to_prime(self, x, y, color=0):
        # Secret decoder ring:
        #  http://www.sdss.org/dr7/products/general/astrometry.html
        # (color)0 is called riCut;
        # g0, g1, g2, and g3 are called
        #    dRow0, dRow1, dRow2, and dRow3, respectively;
        # h0, h1, h2, and h3 are called
        #    dCol0, dCol1, dCol2, and dCol3, respectively;
        # px and py are called csRow and csCol, respectively;
        # and qx and qy are called ccRow and ccCol, respectively.
        color0 = self._get_ricut()
        g0, g1, g2, g3 = self._get_drow()
        h0, h1, h2, h3 = self._get_dcol()
        px, py, qx, qy = self._get_cscc()

        # #$(%*&^(%$%*& bad documentation.
        (px,py) = (py,px)
        (qx,qy) = (qy,qx)

        yprime = y + g0 + g1 * x + g2 * x**2 + g3 * x**3
        xprime = x + h0 + h1 * x + h2 * x**2 + h3 * x**3

        # The code below implements this, vectorized:
        # if color < color0:
        #   xprime += px * color
        #   yprime += py * color
        # else:
        #   xprime += qx
        #   yprime += qy
        qx = qx * np.ones_like(x)
        qy = qy * np.ones_like(y)
        xprime += np.where(color < color0, px * color, qx)
        yprime += np.where(color < color0, py * color, qy)

        return (xprime, yprime)

    def prime_to_pixel(self, xprime, yprime,  color=0):
        color0 = self._get_ricut()
        g0, g1, g2, g3 = self._get_drow()
        h0, h1, h2, h3 = self._get_dcol()
        px, py, qx, qy = self._get_cscc()

        # #$(%*&^(%$%*& bad documentation.
        (px,py) = (py,px)
        (qx,qy) = (qy,qx)

        qx = qx * np.ones_like(xprime)
        qy = qy * np.ones_like(yprime)
        xprime -= np.where(color < color0, px * color, qx)
        yprime -= np.where(color < color0, py * color, qy)

        # Now invert:
        #   yprime = y + g0 + g1 * x + g2 * x**2 + g3 * x**3
        #   xprime = x + h0 + h1 * x + h2 * x**2 + h3 * x**3
        x = xprime - h0
        # dumb-ass Newton's method
        dx = 1.
        # FIXME -- should just update the ones that aren't zero
        # FIXME -- should put in some failsafe...
        while np.max(np.abs(np.atleast_1d(dx))) > 1e-10:
            xp    = x + h0 + h1 * x + h2 * x**2 + h3 * x**3
            dxpdx = 1 +      h1     + h2 * 2*x +  h3 * 3*x**2
            dx = (xprime - xp) / dxpdx
            x += dx
        y = yprime - (g0 + g1 * x + g2 * x**2 + g3 * x**3)
        return (x, y)

    def radec_to_munu_single_c(self, ra, dec):
        ''' Compute ra,dec to mu,nu for a single RA,Dec, calling C code'''
        mu,nu = cutils.radec_to_munu(ra, dec, self.node, self.incl)
        return mu,nu

    def radec_to_munu(self, ra, dec):
        '''
        RA,Dec in degrees

        mu,nu (great circle coords) in degrees
        '''
        node,incl = self.node, self.incl
        assert(ra is not None)
        assert(dec is not None)
        ra, dec = np.deg2rad(ra), np.deg2rad(dec)
        mu = node + np.arctan2(np.sin(ra - node) * np.cos(dec) * np.cos(incl) +
                               np.sin(dec) * np.sin(incl),
                               np.cos(ra - node) * np.cos(dec))
        nu = np.arcsin(-np.sin(ra - node) * np.cos(dec) * np.sin(incl) +
                       np.sin(dec) * np.cos(incl))
        mu, nu = np.rad2deg(mu), np.rad2deg(nu)
        mu += (360. * (mu < 0))
        mu -= (360. * (mu > 360))
        return (mu, nu)

    def munu_to_radec(self, mu, nu):
        node,incl = self.node, self.incl
        assert(mu is not None)
        assert(nu is not None)
        # just in case you thought we needed *more* rad/deg conversions...
        return munu_to_radec_deg(mu, nu, np.rad2deg(node), np.rad2deg(incl))


if cutils is not None:
    AsTrans.radec_to_munu_single = AsTrans.radec_to_munu_single_c
    AsTrans.radec_to_pixel_single = AsTrans.radec_to_pixel_single_c
else:
    AsTrans.radec_to_munu_single = AsTrans.radec_to_munu
    AsTrans.radec_to_pixel_single = AsTrans.radec_to_pixel_single_py


class TsField(SdssFile):
    def __init__(self, *args, **kwargs):
        super(TsField, self).__init__(*args, **kwargs)
        self.filetype = 'tsField'
        self.exptime = 53.907456
    def setHdus(self, p):
        self.hdus = p
        self.table = fits_table(self.hdus[1].data)[0]
        T = self.table
        self.aa = T.aa.astype(float)
        self.kk = T.kk.astype(float)
        self.airmass = T.airmass

    def getAsTrans(self, band):
        bandi = band_index(band)
        band = band_name(band)
        #node,incl = self.getNode(), self.getIncl()
        hdr = self.hdus[0].header
        node = np.deg2rad(hdr.get('NODE'))
        incl = np.deg2rad(hdr.get('INCL'))
        asTrans = AsTrans(self.run, self.camcol, self.field, band=band,
                          node=node, incl=incl, astrans=self.table)
        return asTrans

    #magL = -(2.5/ln(10))*[asinh((f/f0)/2b)+ln(b)]
    # luptitude == arcsinh mag
    # band: int
    def luptitude_to_counts(self, L, band):
        # from arcsinh softening parameters table
        #   http://www.sdss.org/dr7/algorithms/fluxcal.html#counts2mag
        b = [1.4e-10, 0.9e-10, 1.2e-10, 1.8e-10, 7.4e-10]

        b = b[band]
        maggies = 2.*b * np.sinh(-0.4 * np.log(10.) * L - np.log(b))
        dlogcounts = -0.4 * (self.aa[band] + self.kk[band] * self.airmass[band])
        return (maggies * self.exptime) * 10.**dlogcounts

    def get_zeropoint(self, band):
        return (2.5 * np.log10(self.exptime)
                -(self.aa[band] + self.kk[band] * self.airmass[band]))
        
    # band: int
    def mag_to_counts(self, mag, band):
        # log_10(counts)
        logcounts = (-0.4 * mag + np.log10(self.exptime)
                     - 0.4*(self.aa[band] + self.kk[band] * self.airmass[band]))
        #logcounts = np.minimum(logcounts, 308.)
        #olderrs = np.seterr(all='print')
        rtn = 10.**logcounts
        #np.seterr(**olderrs)
        return rtn

    def counts_to_mag(self, counts, band):
        # http://www.sdss.org/dr5/algorithms/fluxcal.html#counts2mag
        # f/f0 = counts/exptime * 10**0.4*(aa + kk * airmass)
        # mag = -2.5 * log10(f/f0)
        return -2.5 * (np.log10(counts / self.exptime) +
                       0.4 * (self.aa[band] + self.kk[band] * self.airmass[band]))


    
class FpObjc(SdssFile):
    def __init__(self, *args, **kwargs):
        super(FpObjc, self).__init__(*args, **kwargs)
        self.filetype = 'fpObjc'

class FpM(SdssFile):
    def __init__(self, *args, **kwargs):
        super(FpM, self).__init__(*args, **kwargs)
        self.filetype = 'fpM'
        self.maskmap = None

    def setHdus(self, p):
        self.hdus = p

    def getMaskPlane(self, name):
        # Mask planes are described in HDU 11 (the last HDU)
        if self.maskmap is None:
            self.maskmap = {}
            T = fits_table(self.hdus[-1].data)
            T.cut(T.defname == 'S_MASKTYPE')
            for k,v in zip(T.attributename, T.value):
                k = k.replace('S_MASK_', '')
                if k == 'S_NMASK_TYPES':
                    continue
                self.maskmap[k] = v
        if not name in self.maskmap:
            raise RuntimeError('Unknown mask plane \"%s\"' % name)

        data = self.hdus[1 + self.maskmap[name]].data
        try:
            if data.get_nrows() == 0:
                return None
        except:
            pass
        return fits_table(data)

    def setMaskedPixels(self, name, img, val, roi=None):
        M = self.getMaskPlane(name)
        if M is None:
            return
        if roi is not None:
            x0,x1,y0,y1 = roi

        for (c0,c1,r0,r1,coff,roff) in zip(M.cmin,M.cmax,M.rmin,M.rmax,
                                           M.col0, M.row0):
            assert(coff == 0)
            assert(roff == 0)
            if roi is not None:
                (outx,nil) = get_overlapping_region(c0-x0, c1+1-x0, 0, x1-x0)
                (outy,nil) = get_overlapping_region(r0-y0, r1+1-y0, 0, y1-y0)
                img[outy,outx] = val
            else:
                img[r0:r1+1, c0:c1+1] = val
        

class FpC(SdssFile):
    def __init__(self, *args, **kwargs):
        super(FpC, self).__init__(*args, **kwargs)
        self.filetype = 'fpC'
    def getImage(self):
        return self.image
    def getHeader(self):
        return self.header

class PsField(SdssFile):
    def __init__(self, *args, **kwargs):
        super(PsField, self).__init__(*args, **kwargs)
        self.filetype = 'psField'

    def setHdus(self, p):
        self.hdus = p
        t = fits_table(p[6].data)
        # the table has only one row...
        assert(len(t) == 1)
        t = t[0]
        #self.table = t
        self.gain = t.gain
        self.dark_variance = t.dark_variance
        self.sky = t.sky
        self.skyerr = t.skyerr
        self.psp_status = t.status
        # Double-Gaussian PSF params
        self.dgpsf_s1 = t.psf_sigma1_2g
        self.dgpsf_s2 = t.psf_sigma2_2g
        self.dgpsf_b  = t.psf_b_2g
        # summary PSF width (sigmas)
        self.psf_fwhm = t.psf_width * (2.*np.sqrt(2.*np.log(2.)))

        # 2-gaussian plus power-law PSF params
        self.plpsf_s1 = t.psf_sigma1
        self.plpsf_s2 = t.psf_sigma2
        self.plpsf_b = t.psf_b
        self.plpsf_p0 = t.psf_p0
        self.plpsf_beta = t.psf_beta
        self.plpsf_sigmap = t.psf_sigmap


        t = fits_table(p[8].data)
        self.per_run_apcorrs = t.ap_corr_run

    def getPowerLaw(self, bandnum):
        ''' Returns:

        (a1, sigma_1,
        a2, sigma_2,
        a3, sigma_power, beta_power)

        Where a1 is the amplitude of the first Gaussian and sigma_1 is
        its standard deviation; a2 and sigma_2 are the same for the
        second Gaussian component, and a3 is the amplitude for the
        power-law component.  Sigma is the scale length, beta the
        power.

        RHL claims:
          func = a*[exp(-x^2/(2*sigmax1^2) - y^2/(2*sigmay1^2)) +
                    b*exp(-x^2/(2*sigmax2^2) - y^2/(2*sigmay2^2)) +
                    p0*(1 + r^2/(beta*sigmap^2))^{-beta/2}]

        '''
        return (1., self.plpsf_s1[bandnum],
                self.plpsf_b[bandnum], self.plpsf_s1[bandnum],
                self.plpsf_p0[bandnum], self.plpsf_sigmap[bandnum],
                self.plpsf_beta[bandnum])

    def getPsfFwhm(self, bandnum):
        return self.psf_fwhm[bandnum]

    def getDoubleGaussian(self, bandnum, normalize=False):
        # http://www.sdss.org/dr7/dm/flatFiles/psField.html
        # good = PSP_FIELD_OK
        status = self.psp_status[bandnum]
        if status != 0:
            print('Warning: PsField status[band=%s] =' % (bandnum), status)

        # b is the "ratio of G2 to G1 at the origin", ie, not the
        # straight Gaussian amplitudes
        a  = 1.0
        s1 = self.dgpsf_s1[bandnum]
        s2 = self.dgpsf_s2[bandnum]
        b  = self.dgpsf_b[bandnum]

        # value at center is 1./(2.*pi*sigma**2)

        if normalize:
            b *= (s2/s1)**2
            absum = (a + b)
            a /= absum
            b /= absum
        
        return (float(a), float(s1), float(b), float(s2))
 
    def getEigenPsfs(self, bandnum):
        '''
        Returns a numpy array of shape, eg, (4, 51, 51).
        '''
        T = fits_table(self.hdus[bandnum+1].data)
        psfs = []
        for psf,h,w in zip(T.rrows, T.rnrow, T.rncol):
            psfs.append(psf.reshape((h,w)))
        psfs = np.array(psfs)
        return psfs

    def getEigenPolynomials(self, bandnum):
        '''
        Returns [ (xorder, yorder, coeffs),  (xorder, yorder, coeffs), ...]
        one tuple per eigen-PSF.
        xorder and yorder are np arrays of integers
        coeffs is a numpy array of floating-point coefficients
        '''
        T = fits_table(self.hdus[bandnum+1].data)
        terms = []
        for k in range(len(T)):
            nrb = T.nrow_b[k]
            ncb = T.ncol_b[k]
            c = T.c[k]
            # !!!
            c = c.copy()
            c = c.reshape(5, 5)
            c = c[:nrb,:ncb]
            (gridc,gridr) = np.meshgrid(np.arange(ncb), np.arange(nrb))
            # remove the 1e-3 coordinate prescaling
            c *= (1e-3 ** (gridr + gridc))
            I = np.flatnonzero(c)
            terms.append((gridr.flat[I], gridc.flat[I], c.flat[I]))
        return terms

    def correlateEigenPsf(self, bandnum, img):
        from scipy.ndimage.filters import correlate

        eigenpsfs = self.getEigenPsfs(bandnum)
        eigenterms = self.getEigenPolynomials(bandnum)
        H,W = img.shape
        corr = np.zeros((H,W))
        xx,yy = np.arange(W).astype(float), np.arange(H).astype(float)
        for epsf, (XO,YO,C) in zip(eigenpsfs, eigenterms):
            k = reduce(np.add, [np.outer(yy**yo, xx**xo) * c
                                for xo,yo,c in zip(XO,YO,C)])
            assert(k.shape == img.shape)
            # Trim symmetric zero-padding off the epsf.
            # This will fail spectacularly given an all-zero eigen-component.
            while True:
                H,W = epsf.shape
                if (np.all(epsf[:,0] == 0) and np.all(epsf[:,-1] == 0) and
                    np.all(epsf[0,:] == 0) and np.all(epsf[-1,:] == 0)):
                    # Trim!
                    epsf = epsf[1:-1, 1:-1]
                else:
                    break
            corr += k * correlate(img, epsf)
        return corr

    def getPsfAtPoints(self, bandnum, x, y):
        '''
        Reconstruct the SDSS model PSF from KL basis functions.

        x,y can be scalars or 1-d numpy arrays.

        Return value:
        if x,y are scalars: a PSF image
        if x,y are arrays:  a list of PSF images
        '''
        rtnscalar = np.isscalar(x) and np.isscalar(y)
        x = np.atleast_1d(x).astype(float)
        y = np.atleast_1d(y).astype(float)

        eigenpsfs = self.getEigenPsfs(bandnum)
        eigenpolys = self.getEigenPolynomials(bandnum)

        # From the IDL docs:
        # http://photo.astro.princeton.edu/photoop_doc.html#SDSS_PSF_RECON
        #   acoeff_k = SUM_i{ SUM_j{ (0.001*ROWC)^i * (0.001*COLC)^j * C_k_ij } }
        #   psfimage = SUM_k{ acoeff_k * RROWS_k }

        # we assume all the eigen-psfs are the same size.
        assert(len(np.unique([psf.shape for psf in eigenpsfs])) == 1)

        xx,yy = np.broadcast_arrays(x, y)
        N = len(xx.flat)
        psfimgs = np.zeros((N,) + eigenpsfs[0].shape)
        for epsf, (XO, YO, C) in zip(eigenpsfs, eigenpolys):
            kk = reduce(np.add, [(xx.flat ** xo) * (yy.flat ** yo) * c
                                 for (xo,yo,c) in zip(XO,YO,C)])
            psfimgs += epsf[np.newaxis,:,:] * kk[:,np.newaxis,np.newaxis]

        if rtnscalar:
            return psfimgs[0,:,:]
        # convert back to a list...
        return [psfimgs[i,:,:] for i in range(N)]

    def getGain(self, band=None):
        if band is not None:
            return self.gain[band]
        return self.gain

    def getDarkVariance(self, band=None):
        if band is not None:
            return self.dark_variance[band]
        return self.dark_variance

    def getSky(self, band=None):
        if band is not None:
            return self.sky[band]
        return self.sky

    def getSkyErr(self, band=None):
        if band is not None:
            return self.skyerr[band]
        return self.skyerr





#! /usr/bin/env python3
# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
from __future__ import print_function
import sys
from optparse import OptionParser
from astrometry.util.util import *

if __name__ == '__main__':
    parser = OptionParser('usage: %prog [options] <outfn>')
    parser.add_option('-r', '--ra', dest='ra', type='float', help='RA (deg)')
    parser.add_option('-d', '--dec', dest='dec', type='float', help='Dec (deg)')
    parser.add_option('-s', '--size', dest='size', type='float', help='Field width (deg)')
    parser.add_option('-p', '--pixscale', type='float', help='Pixel scale (arcsec/pixel)')
    parser.add_option('-W', '--width', dest='w', type='int', help='Image width, default %default', default=1024)
    parser.add_option('-H', '--height', dest='h', type='int', help='Image height, default %default', default=1024)
    parser.set_defaults(ra=None, dec=None, size=None, w=None, h=None)
    opt,args = parser.parse_args()

    if len(args) == 0:
        parser.print_help()
        sys.exit(0)

    if opt.pixscale is None and opt.size is None:
        print('Must specify --pixscale or --size')
        parser.print_help()
        sys.exit(-1)
    pixscale = 0.
    if opt.size is not None:
        pixscale = opt.size / opt.w
    elif opt.pixscale is not None:
        pixscale = opt.pixscale / 3600.

    if opt.ra is None or opt.dec is None:
        print('Must specify --ra and --dec')
        parser.print_help()
        sys.exit(-1)

    wcs = Tan(*[float(x) for x in [
        opt.ra, opt.dec, 0.5 + (opt.w / 2.), 0.5 + (opt.h / 2.),
        -pixscale, 0., 0., pixscale, opt.w, opt.h,
    ]])

    wcs.write_to(args[0])



# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
from math import pi
import numpy as np

def estimate_mode(img, lo=None, hi=None, plo=1, phi=70, bins1=30,
                   flo=0.5, fhi=0.8, bins2=30,
                   return_fit=False, raiseOnWarn=False):
    # Estimate sky level by: compute the histogram within [lo,hi], fit
    # a parabola to the log-counts, return the argmax of that parabola.

    # Coarse bin to find the peak (mode)
    if lo is None:
        lo = np.percentile(img, plo)
    if hi is None:
        hi = np.percentile(img, phi)

    binedges1 = np.linspace(lo, hi, bins1+1)
    counts1,e = np.histogram(img.ravel(), bins=binedges1)
    bincenters1 = binedges1[:-1] + (binedges1[1]-binedges1[0])/2.
    maxbin = np.argmax(counts1)
    maxcount = counts1[maxbin]
    mode = bincenters1[maxbin]

    # Search for bin containing < {flo,fhi} * maxcount
    ilo = maxbin
    while ilo > 0:
        ilo -= 1
        if counts1[ilo] < flo*maxcount:
            break
    ihi = maxbin
    while ihi < bins1-1:
        ihi += 1
        if counts1[ihi] < fhi*maxcount:
            break

    lo = bincenters1[ilo]
    hi = bincenters1[ihi]

    binedges = np.linspace(lo, hi, bins2)
    counts,e = np.histogram(img.ravel(), bins=binedges)
    bincenters = binedges[:-1] + (binedges[1]-binedges[0])/2.

    b = np.log10(np.maximum(1, counts))

    xscale = 0.5 * (hi - lo)
    x0 = (hi + lo) / 2.
    x = (bincenters - x0) / xscale

    A = np.zeros((len(x), 3))
    A[:,0] = 1.
    A[:,1] = x
    A[:,2] = x**2
    res = np.linalg.lstsq(A, b, rcond=None)
    X = res[0]
    mx = -X[1] / (2. * X[2])
    mx = (mx * xscale) + x0

    warn = None

    if not (mx > lo and mx < hi):
        if raiseOnWarn:
            raise ValueError('sky estimate not bracketed by peak: lo %f, sky %f, hi %f' % (lo, mx, hi))
        warn = 'WARNING: sky estimate not bracketed by peak: lo %f, sky %f, hi %f' % (lo, mx, hi)

    if return_fit:
        bfit = X[0] + X[1] * x + X[2] * x**2
        return (x * xscale + x0, b, bfit, mx, warn, binedges1,counts1)

    return mx


def parse_ranges(s):
    '''
    Parse PBS job array-style ranges: NNN,MMM-NNN,PPP

    *s*: string

    Returns: [ int, int, ... ]
    '''
    tiles = []
    words = s.split()
    for w in words:
        for a in w.split(','):
            if '-' in a:
                aa = a.split('-')
                if len(aa) != 2:
                    raise RuntimeError('With an arg containing a dash, expect two parts, in word "%s"' % a)
                start = int(aa[0])
                end = int(aa[1])
                for i in range(start, end+1):
                    tiles.append(i)
            else:
                tiles.append(int(a))
    return tiles


def patch_image(img, mask, dxdy = [(-1,0),(1,0),(0,-1),(0,1)],
                required=None):
    '''
    Patch masked pixels by iteratively averaging non-masked neighboring pixels.

    WARNING: this modifies BOTH the "img" and "mask" arrays!

    mask: True for good pixels
    required: if non-None: True for pixels you want to be patched.
    dxdy: Pixels to average in, relative to pixels to be patched.

    Returns True if patching was successful.
    '''
    assert(img.shape == mask.shape)
    assert(len(img.shape) == 2)
    h,w = img.shape
    Nlast = -1
    while True:
        needpatching = np.logical_not(mask)
        if required is not None:
            needpatching *= required
        I = np.flatnonzero(needpatching)
        if len(I) == 0:
            break
        if len(I) == Nlast:
            return False
        #print 'Patching', len(I), 'pixels'
        Nlast = len(I)
        iy,ix = np.unravel_index(I, img.shape)
        psum = np.zeros(len(I), img.dtype)
        pn = np.zeros(len(I), int)

        for dx,dy in dxdy:
            ok = True
            if dx < 0:
                ok = ok * (ix >= (-dx))
            if dx > 0:
                ok = ok * (ix <= (w-1-dx))
            if dy < 0:
                ok = ok * (iy >= (-dy))
            if dy > 0:
                ok = ok * (iy <= (h-1-dy))

            # darn, NaN * False = NaN, not zero.
            finite = np.isfinite(img [iy[ok]+dy, ix[ok]+dx])
            ok[ok] *= finite

            psum[ok] += (img [iy[ok]+dy, ix[ok]+dx] *
                         mask[iy[ok]+dy, ix[ok]+dx])
            pn[ok] += mask[iy[ok]+dy, ix[ok]+dx]

            # print 'ix', ix
            # print 'iy', iy
            # print 'dx,dy', dx,dy
            # print 'ok', ok
            # print 'psum', psum
            # print 'pn', pn

        img.flat[I] = (psum / np.maximum(pn, 1)).astype(img.dtype)
        mask.flat[I] = (pn > 0)
        #print 'Patched', np.sum(pn > 0)
    return True

def clip_wcs(wcs1, wcs2, makeConvex=True, pix1=None, pix2=None):
    '''
    Returns a pixel-space polygon in WCS1 after it is clipped by WCS2.
    Returns an empty list if the two WCS headers do not intersect.

    Note that due to weakness in the clip_polygon method, wcs2 must be convex.
    If makeConvex=True, we find the convex hull and clip with that.

    If pix1 is not None, pix1=(xx,yy), xx and yy lists of pixel coordinates defining
    the boundary of the image, in CLOCKWISE order; default is the edges and midpoints.

    '''
    if pix1 is None:
        w1,h1 = wcs1.get_width(), wcs1.get_height()
        x1,x2,x3 = 0.5, w1/2., w1+0.5
        y1,y2,y3 = 0.5, h1/2., h1+0.5
        xx = [x1, x1, x1, x2, x3, x3, x3, x2]
        yy = [y1, y2, y3, y3, y3, y2, y1, y1]
    else:
        xx,yy = pix1
    #rr,dd = wcs1.pixelxy2radec(xx, yy)

    if pix2 is None:
        w2,h2 = wcs2.get_width(), wcs2.get_height()
        x1,x2,x3 = 0.5, w2/2., w2+0.5
        y1,y2,y3 = 0.5, h2/2., h2+0.5
        XX = [x1, x1, x1, x2, x3, x3, x3, x2]
        YY = [y1, y2, y3, y3, y3, y2, y1, y1]
    else:
        XX,YY = pix2
    rr,dd = wcs2.pixelxy2radec(XX, YY)
    ok,XX,YY = wcs1.radec2pixelxy(rr, dd)
    # XX,YY is the clip polygon in wcs1 pixel coords.
    # Not necessarily clockwise at this point!

    #print 'XX,YY', XX, YY

    if makeConvex:
        from scipy.spatial import ConvexHull
        points = np.vstack((XX,YY)).T
        #print 'points', points.shape
        hull = ConvexHull(points)
        #print 'Convex hull:', hull
        #print hull.vertices

        # ConvexHull returns the vertices in COUNTER-clockwise order.  Reverse.
        v = np.array(list(reversed(hull.vertices))).astype(int)
        XX = XX[v]
        YY = YY[v]

        #plt.plot(XX, YY, 'm-')
        #plt.plot(XX[0], YY[0], 'mo')
        #plt.savefig('clip2.png')
    else:
        # Ensure points are listed in CLOCKWISE order.
        crosses = []
        for i in range(len(XX)):
            j = (i + 1) % len(XX)
            k = (i + 2) % len(XX)
            xi,yi = XX[i], YY[i]
            xj,yj = XX[j], YY[j]
            xk,yk = XX[k], YY[k]
            dx1, dy1 = xj - xi, yj - yi
            dx2, dy2 = xk - xj, yk - yj
            cross = dx1 * dy2 - dy1 * dx2
            #print 'cross', cross
            crosses.append(cross)
        crosses = np.array(crosses)
        #print 'cross products', crosses
        if np.all(crosses >= 0):
            # Reverse
            #print 'Reversing wcs2 points'
            XX = np.array(list(reversed(XX)))
            YY = np.array(list(reversed(YY)))

    clip = clip_polygon(list(zip(xx, yy)), list(zip(XX, YY)))
    clip = np.array(clip)

    if False:
        import pylab as plt
        plt.clf()
        plt.plot(xx, yy, 'b.-')
        plt.plot(xx[0], yy[0], 'bo')
        plt.plot(XX, YY, 'r.-')
        plt.plot(XX[0], YY[0], 'ro')
        if len(clip) > 0:
            plt.plot(clip[:,0], clip[:,1], 'm.-', alpha=0.5, lw=2)
        plt.savefig('clip1.png')

    return clip



def polygon_area(poly):
    '''
    NOTE, unlike many of the other methods in this module, takes:

    poly = (xx,yy)

    where xx,yy MUST repeat the starting point at the end of the polygon.
    '''
    xx,yy = poly
    x,y = np.mean(xx), np.mean(yy)
    area = 0.
    for dx0,dy0,dx1,dy1 in zip(xx-x, yy-y, xx[1:]-x, yy[1:]-y):
        # area: 1/2 cross product
        area += np.abs(dx0 * dy1 - dx1 * dy0)
    return 0.5 * area

def clip_polygon(poly1, poly2):
    '''
    Returns a new polygon resulting from taking poly1 and clipping it
    to lie inside poly2.

    WARNING, the polygons must be listed in CLOCKWISE order.

    WARNING, the clipping polygon, poly2, must be CONVEX.
    '''
    # from clipper import Clipper, Point, PolyType, ClipType, PolyFillType
    # '''
    # '''
    # c = Clipper()
    # p1 = [Point(x,y) for x,y in poly1]
    # p2 = [Point(x,y) for x,y in poly2]
    # c.AddPolygon(p1, PolyType.Subject)
    # c.AddPolygon(p2, PolyType.Clip)
    # solution = []
    # pft = PolyFillType.EvenOdd
    # result = c.Execute(ClipType.Intersection, solution, pft, pft)
    # if len(solution) > 1:
    #     raise RuntimeError('Polygon clipping results in non-simple polygon')
    # assert(result)
    # #print 'Result:', result
    # #print 'Solution:', solution
    # return [(s.x, s.y) for s in solution[0]]

    # Sutherland-Hodgman algorithm -- thanks, Wikipedia!
    N2 = len(poly2)
    # clip by each edge in turn.
    for j in range(N2):
        # target "left_right" value
        clip1 = poly2[j]
        clip2 = poly2[(j+1)%N2]
        LRinside = _left_right(clip1, clip2, poly2[(j+2)%N2])
        # are poly vertices inside or outside the clip polygon?
        isinside = [_left_right(clip1, clip2, p) == LRinside
                    for p in poly1]
        # the resulting clipped polygon
        clipped = []
        N1 = len(poly1)
        for i in range(N1):
            S = poly1[i]
            E = poly1[(i+1)%N1]
            Sin = isinside[i]
            Ein = isinside[(i+1)%N1]
            if Ein:
                if not Sin:
                    clipped.append(line_intersection(clip1, clip2, S, E))
                clipped.append(E)
            else:
                if Sin:
                    clipped.append(line_intersection(clip1, clip2, S, E))
        poly1 = clipped
    return poly1


def polygons_intersect(poly1, poly2):
    '''
    Determines whether the given 2-D polygons intersect.

    poly1, poly2: np arrays with shape (N,2)
    '''

    # Check whether any points in poly1 are inside poly2,
    # or vice versa.
    for (px,py) in poly1:
        if point_in_poly(px,py, poly2):
            return (px,py)
    for (px,py) in poly2:
        if point_in_poly(px,py, poly1):
            return (px,py)

    # Check for intersections between line segments.  O(n^2) brutish
    N1 = len(poly1)
    N2 = len(poly2)

    for i in range(N1):
        for j in range(N2):
            xy = line_segments_intersect(poly1[i % N1, :], poly1[(i+1) % N1, :],
                                         poly2[j % N2, :], poly2[(j+1) % N2, :])
            if xy:
                return xy
    return False


def line_segments_intersect(xy1, xy2, xy3, xy4):
    '''
    Determines whether the two given line segments intersect;

    (x1,y1) to (x2,y2)
    and
    (x3,y3) to (x4,y4)
    '''
    (x1,y1) = xy1
    (x2,y2) = xy2
    (x3,y3) = xy3
    (x4,y4) = xy4
    x,y = line_intersection((x1,y1),(x2,y2),(x3,y3),(x4,y4))
    if x is None:
        # Parallel lines
        return False
    if x1 == x2:
        p1,p2 = y1,y2
        p = y
    else:
        p1,p2 = x1,x2
        p = x

    if not ((p >= min(p1,p2)) and (p <= max(p1,p2))):
        return False

    if x3 == x4:
        p1,p2 = y3,y4
        p = y
    else:
        p1,p2 = x3,x4
        p = x

    if not ((p >= min(p1,p2)) and (p <= max(p1,p2))):
        return False
    return (x,y)


def line_intersection(xy1, xy2, xy3, xy4):
    '''
    Determines the point where the lines described by
    (x1,y1) to (x2,y2)
    and
    (x3,y3) to (x4,y4)
    intersect.

    Note that this may be beyond the endpoints of the line segments.

    Probably raises an exception if the lines are parallel, or does
    something numerically crazy.
    '''
    (x1,y1) = xy1
    (x2,y2) = xy2
    (x3,y3) = xy3
    (x4,y4) = xy4
    # This code started with the equation from Wikipedia,
    # then I added special-case handling.
    # bottom = ((x1 - x2) * (y3 - y4) - (y1 - y2) * (x3 - x4))
    # if bottom == 0:
    #   raise RuntimeError("divide by zero")
    # t1 = (x1 * y2 - y1 * x2)
    # t2 = (x3 * y4 - y3 * x4)
    # px = (t1 * (x3 - x4) - t2 * (x1 - x2)) / bottom
    # py = (t1 * (y3 - y4) - t2 * (y1 - y2)) / bottom

    # From http://wiki.processing.org/w/Line-Line_intersection
    bx = float(x2) - float(x1)
    by = float(y2) - float(y1)
    dx = float(x4) - float(x3)
    dy = float(y4) - float(y3)
    b_dot_d_perp = bx*dy - by*dx
    if b_dot_d_perp == 0:
        return None,None
    cx = float(x3) - float(x1)
    cy = float(y3) - float(y1)
    t = (cx*dy - cy*dx) / b_dot_d_perp
    return x1 + t*bx, y1 + t*by

def _left_right(xy1, xy2, xy3):
    '''
    is (x3,y3) to the 'left' or 'right' of the line from (x1,y1) to (x2,y2) ?
    '''
    (x1,y1) = xy1
    (x2,y2) = xy2
    (x3,y3) = xy3
    dx2,dy2 = x2-x1, y2-y1
    dx3,dy3 = x3-x1, y3-y1
    return (dx2 * dy3 - dx3 * dy2) > 0


def point_in_poly(x, y, poly):
    '''
    Performs a point-in-polygon test for numpy arrays of *x* and *y*
    values, and a polygon described as 2-d numpy array (with shape (N,2))

    poly: N x 2 array

    Returns a numpy array of bools.
    '''
    x = np.atleast_1d(x)
    y = np.atleast_1d(y)
    inside = np.zeros(x.shape, bool)
    # This does a winding test -- count how many times a horizontal ray
    # from (-inf,y) to (x,y) crosses the boundary.
    for i in range(len(poly)):
        j = (i-1 + len(poly)) % len(poly)
        xi,xj = poly[i,0], poly[j,0]
        yi,yj = poly[i,1], poly[j,1]

        if yi == yj:
            continue

        I = np.logical_and(
            np.logical_or(np.logical_and(yi <= y, y < yj),
                          np.logical_and(yj <= y, y < yi)),
            x < (xi + ((xj - xi) * (y - yi) / (yj - yi))))
        inside[I] = np.logical_not(inside[I])
    return inside

def lanczos_filter(order, x, out=None):
    x = np.atleast_1d(x)
    nz = np.logical_and(x != 0., np.logical_and(x < order, x > -order))
    nz = np.flatnonzero(nz)
    if out is None:
        out = np.zeros(x.shape, dtype=float)
    else:
        out[x <= -order] = 0.
        out[x >=  order] = 0.
    pinz = pi * x.flat[nz]
    out.flat[nz] = order * np.sin(pinz) * np.sin(pinz / order) / (pinz**2)
    out[x == 0] = 1.
    return out

def get_overlapping_region(xlo, xhi, xmin, xmax):
    '''
    Given a range of integer coordinates that you want to, eg, cut out
    of an image, [xlo, xhi], and bounds for the image [xmin, xmax],
    returns the range of coordinates that are in-bounds, and the
    corresponding region within the desired cutout.

    For example, say you have an image of shape H,W and you want to
    cut out a region of halfsize "hs" around pixel coordinate x,y, but
    so that coordinate x,y is centered in the cutout even if x,y is
    close to the edge.  You can do:

    cutout = np.zeros((hs*2+1, hs*2+1), img.dtype)
    iny,outy = get_overlapping_region(y-hs, y+hs, 0, H-1)
    inx,outx = get_overlapping_region(x-hs, x+hs, 0, W-1)
    cutout[outy,outx] = img[iny,inx]
    
    '''
    if xlo > xmax or xhi < xmin or xlo > xhi or xmin > xmax:
        return ([], [])

    assert(xlo <= xhi)
    assert(xmin <= xmax)

    xloclamp = max(xlo, xmin)
    Xlo = xloclamp - xlo

    xhiclamp = min(xhi, xmax)
    Xhi = Xlo + (xhiclamp - xloclamp)

    #print 'xlo, xloclamp, xhiclamp, xhi', xlo, xloclamp, xhiclamp, xhi
    assert(xloclamp >= xlo)
    assert(xloclamp >= xmin)
    assert(xloclamp <= xmax)
    assert(xhiclamp <= xhi)
    assert(xhiclamp >= xmin)
    assert(xhiclamp <= xmax)
    #print 'Xlo, Xhi, (xmax-xmin)', Xlo, Xhi, xmax-xmin
    assert(Xlo >= 0)
    assert(Xhi >= 0)
    assert(Xlo <= (xhi-xlo))
    assert(Xhi <= (xhi-xlo))

    return (slice(xloclamp, xhiclamp+1), slice(Xlo, Xhi+1))



if __name__ == '__main__':
    import matplotlib
    matplotlib.use('Agg')
    import pylab as plt
    import numpy as np
    from astrometry.util.plotutils import *
    ps = PlotSequence('miscutils')

    np.random.seed(42)

    if True:
        p2 = np.array([[0,0],[0,4],[4,4],[4,0]])

        for i,p1 in enumerate([ np.array([[0,0],[0,2],[2,2],[2,0]]),
                                np.array([[-1,-1],[0,2],[2,2],[2,0]]),
                                np.array([[4,0],[0,4],[-4,0],[0,-4]]),
                                np.array([[-1,2],[2,5],[5,2],[2,-1]]),
                                ] + [None]*10):
            if p1 is None:
                p1 = np.random.uniform(high=6., low=-2, size=(4,2))
            pc = np.array(clip_polygon(p1, p2))
            plt.clf()
            I = np.array([0,1,2,3,0])
            plt.plot(p1[I,0], p1[I,1], 'b-', lw=3, alpha=0.5)
            plt.plot(p2[I,0], p2[I,1], 'k-')
            I = np.array(range(len(pc)) + [0])
            plt.plot(pc[I,0], pc[I,1], 'r-')
            plt.axis([-1,5,-1,5])
            plt.savefig('clip-%02i.png' % i)
        import sys
        sys.exit(0)

    if True:
        for i in range(20):
            if i <= 10:
                xy1 = np.array([[0,0],[0,4],[4,4],[4,0]])
            else:
                xy1 = np.random.uniform(high=10., size=(4,2))
            xy2 = np.random.uniform(high=10., size=(4,2))
            plt.clf()
            I = np.array([0,1,2,3,0])
            xy = polygons_intersect(xy1, xy2)
            if xy:
                cc = 'r'
                x,y = xy
                plt.plot(x,y, 'mo', mec='m', mfc='none', ms=20, mew=3, zorder=30)
            else:
                cc = 'k'
            plt.plot(xy1[I,0], xy1[I,1], '-', color=cc, zorder=20, lw=3)
            plt.plot(xy2[I,0], xy2[I,1], '-', color=cc, zorder=20, lw=3)
            ax = plt.axis()
            plt.axis([ax[0]-0.5, ax[1]+0.5, ax[2]-0.5, ax[3]+0.5])
            ps.savefig()

    if False:
        X,Y = np.meshgrid(np.linspace(-1,11, 20), np.linspace(-1,11, 23))
        X = X.ravel()
        Y = Y.ravel()
        for i in range(20):
            if i == 0:
                xy = np.array([[0,0],[0,10],[10,10],[10,0]])
            else:
                xy = np.random.uniform(high=10., size=(4,2))
            plt.clf()
            I = np.array([0,1,2,3,0])
            plt.plot(xy[I,0], xy[I,1], 'r-', zorder=20, lw=3)
            inside = point_in_poly(X, Y, xy)
            plt.plot(X[inside], Y[inside], 'bo')
            out = np.logical_not(inside)
            plt.plot(X[out], Y[out], 'ro')
            ax = plt.axis()
            plt.axis([ax[0]-0.5, ax[1]+0.5, ax[2]-0.5, ax[3]+0.5])
            ps.savefig()



    if True:
        # intersection()
        for i in range(20):
            if i == 0:
                x1 = x2 = 0
                y1 = 0
                y2 = 1
                x3 = 1
                x4 = -1
                y3 = 0
                y4 = 1
            elif i == 1:
                x1,y1 = 0,0
                x2,y2 = 0,1
                x3,y3 = -3,0
                x4,y4 = -2,0
            elif i == 2:
                x1,y1 = 1,0
                x2,y2 = 0,1
                x3,y3 = -3,0
                x4,y4 = -2,0
            elif i == 3:
                x1,y1 = 0,1
                x2,y2 = 1,0
                x3,y3 = 0,-3
                x4,y4 = 0,-2
            elif i == 4:
                x1,y1 = 0,0
                x2,y2 = 0,1
                x3,y3 = 0,2
                x4,y4 = 0,3
            elif i == 5:
                x1,y1 = -1,0
                x2,y2 = 1, 0
                x3,y3 = 0, 2
                x4,y4 = 0.5, 1
            else:
                xy = np.random.uniform(high=10., size=(8,))
                x1,y1,x2,y2,x3,y3,x4,y4 = xy
            plt.clf()
            plt.plot([x1,x2],[y1,y2], 'r-', zorder=20, lw=3)
            plt.plot([x3,x4],[y3,y4], 'b-', zorder=20, lw=3)
            x,y = line_intersection((x1,y1),(x2,y2),(x3,y3),(x4,y4))
            plt.plot(x, y, 'kx', ms=20, zorder=25)
            plt.plot([x1,x],[y1,y], 'k--', alpha=0.5, zorder=15)
            plt.plot([x2,x],[y2,y], 'k--', alpha=0.5, zorder=15)
            plt.plot([x3,x],[y3,y], 'k--', alpha=0.5, zorder=15)
            plt.plot([x4,x],[y4,y], 'k--', alpha=0.5, zorder=15)

            # line_segments_intersect()
            if line_segments_intersect((x1,y1),(x2,y2),(x3,y3),(x4,y4)):
                plt.plot(x,y, 'mo', mec='m', mfc='none', ms=20, mew=3, zorder=30)
            ax = plt.axis()
            plt.axis([ax[0]-0.5, ax[1]+0.5, ax[2]-0.5, ax[3]+0.5])
            ps.savefig()



# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
from __future__ import print_function
import os
import os.path
import re
import socket
import sys

from ftplib import FTP


class crawler(object):
    dirstack = [ '' ]
    currentdir = None

    ffiles = None
    fdirs = None
    fnomatches = None

    def __init__(self):
        self.ffiles = open('files', 'ab')
        self.fdirs = open('dirs', 'ab')
        self.fnomatches = open('nomatch', 'ab')

    ire = re.compile(r'^(?P<dir>.)' + # 'd' or '-'
                     r'.{9}' + # mode
                     r'\s*\w*' + # space, link count
                     r'\s*\w*' + # space, owner
                     r'\s*\w*' + # space, group
                     r'\s*(?P<size>\d*)' + # filesize
                     r'\s*(?P<month>\w*)' + # space, month
                     r'\s*(?P<day>\w*)' + # space, day
                     r'\s*(?P<time>[\w:]*)' + # space, time
                     r'\s*(?P<name>[\w.~_+-]*)' + # space, name
                     r'$')

    def close(self):
        self.ffiles.close()
        self.fdirs.close()
        self.fnomatches.close()

    def set_dirstack(self, stack):
        self.dirstack = stack

    def write_stack(self):
        f = open('dirstack.tmp', 'wb')
        for d in self.dirstack:
            f.write(d + '\n')
        f.close()
        os.rename('dirstack.tmp', 'dirstack')

    def add_item(self, s):
        print('item', s)
        m = self.ire.match(s)
        if not m:
            print('no match')
            self.fnomatches.write(self.currentdir + ' ' + s + '\n')
            self.fnomatches.flush()
            return

        d = m.group('dir')
        name = m.group('name')
        path = self.currentdir + '/' + name
        
        if d == 'd':
            self.dirstack.append(path)
            self.fdirs.write(path + '\n')
            self.fdirs.flush()
        else:
            self.ffiles.write(path + '\n')
            self.ffiles.flush()



if __name__ == '__main__':
    ftp = None

    crawl = crawler()

    if os.path.exists('dirstack'):
        f = open('dirstack', 'rb')
        stack = f.read().strip().split('\n')
        #stack = []
        #for ln in f:
        #    stack.append(ln)
        print('Dirstack:')
        for d in stack:
            print(d)
        print('(end dirstack)')
        crawl.set_dirstack(stack)

    nrequests = 0
    socket.setdefaulttimeout(10)

    while len(crawl.dirstack):
        if not ftp: # or not (nrequests % 100):
            if ftp:
                print('closing connection.')
                ftp.quit()
            print('opening connection')
            sys.stdout.flush()
            ftp = FTP('galex.stsci.edu')
            ftp.login('anonymous', 'dstn@cs.toronto.edu')
            #ftp.set_debuglevel(2)

        d = crawl.dirstack.pop()
        crawl.currentdir = d
        print('listing "%s"' % d)
        sys.stdout.flush()
        try:
            ftp.dir(d, crawl.add_item)
            crawl.write_stack()
            nrequests += 1
        except Exception as e:
            print('caught exception:', e)
            sys.stdout.flush()
            crawl.dirstack.append(d)
            ftp.close()
            ftp = None

